@@ -1,7 +1,8 @@
-xmlRoot.XMLInternalDocument = 
-function(x, skip = TRUE, ...)
+#xmlRoot.HTMLInternalDocument =
+  xmlRoot.XMLInternalDocument = 
+function(x, skip = TRUE, addFinalizer = NA, ...)
 {
-  .Call("R_xmlRootNode", x, as.logical(skip), PACKAGE = "XML")
+  .Call("R_xmlRootNode", x, as.logical(skip), addFinalizer, PACKAGE = "XML")
 }
 
 setAs("XMLNode", "XMLInternalNode",
@@ -28,9 +29,13 @@ setAs("XMLInternalDocument", "XMLInternalNode",
 
 
 setAs("XMLInternalNode", "XMLInternalDocument",
-        function(from)
-           .Call("R_getXMLNodeDocument", from, PACKAGE = "XML")
-      )
+        function(from) {
+           doc = .Call("R_getXMLNodeDocument", from, PACKAGE = "XML")
+           addDocFinalizer(doc, TRUE)
+           if(is(doc, "HTMLInternalDocument"))
+              class(doc) = c(class(doc), "XMLInternalDocument", "XMLAbstractDocument")
+	   doc
+      })
 
 
 
@@ -63,21 +68,24 @@ function(obj, fun, ...)
 
 asRXMLNode =
 function(node, converters = NULL, trim = TRUE, ignoreBlanks = TRUE)
-   .Call("R_createXMLNode", node, converters, as.logical(trim), as.logical(ignoreBlanks), PACKAGE = "XML")
+{  
+   .Call("R_createXMLNode", node, converters, as.logical(trim), as.logical(ignoreBlanks), PACKAGE = "XML")[[1]]
+}
 
 "[.XMLInternalDocument" =
-function(x, i, j, ..., namespaces = xmlNamespaceDefinitions(x, simplify = TRUE))
+function(x, i, j, ..., namespaces = xmlNamespaceDefinitions(x, simplify = TRUE), addFinalizer = NA)
 {
   if(is.character(i)) {
-    getNodeSet(x, i, ...)
+    getNodeSet(x, i, ..., addFinalizer = addFinalizer)
   } else
      stop("No method for subsetting an XMLInternalDocument with ", class(i))
 }  
 
 "[[.XMLInternalDocument" =
-function(x, i, j, ..., exact = NA, namespaces = xmlNamespaceDefinitions(x, simplify = TRUE))
+function(x, i, j, ..., exact = NA, namespaces = xmlNamespaceDefinitions(x, simplify = TRUE),
+           addFinalizer = NA)
 {
-  ans = x[i]
+  ans = x[i, addFinalizer = addFinalizer]
   if(length(ans) > 1)
     warning(length(ans), " elements in node set. Returning just the first one! (Use [])")
   ans[[1]]
@@ -117,15 +125,34 @@ function(x)
 xmlAttrs.XMLInternalNode = 
 function(node, addNamespacePrefix = FALSE, addNamespaceURLs = TRUE, ...)
 {
-  .Call("RS_XML_xmlNodeAttributes",  node, as.logical(addNamespacePrefix), as.logical(addNamespaceURLs), PACKAGE = "XML")
+  ans = .Call("RS_XML_xmlNodeAttributes",  node, as.logical(addNamespacePrefix), as.logical(addNamespaceURLs), PACKAGE = "XML")
+  if(length(attr(ans, "namespaces")))
+    ans = new("XMLAttributes", ans) # class(ans) = "XMLAttributes"
+  
+  ans
 }
 
+#setOldClass(c("XMLAttributes", "character"))
+setClass("XMLAttributes", contains = "character")
+
+setMethod("show", "XMLAttributes",
+           function(object)
+              print(unclass(object)))
+
+setMethod('[', c('XMLAttributes', "ANY"),
+function(x, i, j, ...)
+{
+  ans = callNextMethod()
+  i = match(i, names(x))
+  structure(ans, namespaces = attr(x, "namespaces")[i], class = class(x))
+})
+
 
 
 xmlChildren.XMLInternalNode =
-function(x, addNames = TRUE, omitNodeTypes = c("XMLXIncludeStartNode", "XMLXIncludeEndNode"), ...)
+function(x, addNames = TRUE, omitNodeTypes = c("XMLXIncludeStartNode", "XMLXIncludeEndNode"), addFinalizer = NA, ...)
 {
-  kids = .Call("RS_XML_xmlNodeChildrenReferences", x, as.logical(addNames), PACKAGE = "XML")
+  kids = .Call("RS_XML_xmlNodeChildrenReferences", x, as.logical(addNames), addFinalizer, PACKAGE = "XML")
   
 if(length(omitNodeTypes))
     kids = kids[! sapply(kids, function(x) any(inherits(x, omitNodeTypes)) )]   
@@ -155,13 +182,23 @@ function(obj)
 
 "[[.XMLInternalNode" <-
 #setMethod("[[", "XMLInternalNode",
-function(x, i, j, ...)
+function(x, i, j, ..., addFinalizer = NA)
 {
   if(inherits(i, "formula")) {
-    return(getNodeSet(x, i, if(missing(j)) character() else j)[[1]])
+    return(getNodeSet(x, i, if(missing(j)) character() else j, addFinalizer = addFinalizer, ...)[[1]])
   }
   
-  kids = xmlChildren(x)
+  if(is.na(i))
+     return(NULL)
+     # Get the individual elements rather than all the children and then subset those
+  return(
+       if(is(i, "numeric"))
+          .Call("R_getChildByIndex", x, as.integer(i), as.logical(addFinalizer), PACKAGE = "XML")
+       else
+          .Call("R_getChildByName", x, as.character(i), as.logical(addFinalizer), PACKAGE = "XML")
+       )
+  
+  kids = xmlChildren(x, addFinalizer = addFinalizer)
   if(length(kids) == 0)
     return(NULL)
   
@@ -177,9 +214,9 @@ function(x, i, j, ...)
 
 
 "[.XMLInternalNode" <-
-function(x, i, j, ...)
+function(x, i, j, ..., addFinalizer = NA)
 {
-  kids = xmlChildren(x)
+  kids = xmlChildren(x, addFinalizer = addFinalizer)
   if(is.logical(i))
     i = which(i)
 
@@ -194,7 +231,7 @@ function(x, i, j, ...)
 
 
 xmlValue.XMLInternalNode =
-function(x, ignoreComments = FALSE, recursive = TRUE, encoding = CE_NATIVE)
+function(x, ignoreComments = FALSE, recursive = TRUE, encoding = getEncoding(x), trim = FALSE) #CE_NATIVE)
 {
 
   encoding = if(is.integer(encoding)) 
@@ -206,14 +243,19 @@ function(x, ignoreComments = FALSE, recursive = TRUE, encoding = CE_NATIVE)
      if(xmlSize(x) == 0)
        return(character())
 
-    i = sapply(xmlChildren(x), inherits, "XMLInternalTextNode")
+    kids = xmlChildren(x, addFinaliizer = FALSE)
+    i = sapply(kids, inherits, "XMLInternalTextNode")
     if(any(i))
-      return(paste(unlist(lapply(xmlChildren(x)[i], xmlValue, ignoreComments, recursive = TRUE, encoding = encoding)), collapse = ""))
+      return(paste(unlist(lapply(kids[i], xmlValue, ignoreComments, recursive = TRUE, encoding = encoding)), collapse = ""))
     else
       return(character())
    }
   
-  .Call("R_xmlNodeValue", x, NULL, encoding, PACKAGE = "XML") # 2nd argument ignored.
+  ans = .Call("R_xmlNodeValue", x, NULL, encoding, PACKAGE = "XML") # 2nd argument ignored.
+  if(trim)
+     trim(ans)
+  else
+     ans
 }
 
 setS3Method("xmlValue", "XMLInternalNode")
@@ -252,49 +294,66 @@ setMethod("xmlValue<-", "XMLAbstractNode",
 
 names.XMLInternalNode =
 function(x)
-  xmlSApply(x, xmlName)
+  xmlSApply(x, xmlName, addFinalizer = FALSE)
 
 xmlApply.XMLInternalNode =
-function(X, FUN, ..., omitNodeTypes = c("XMLXIncludeStartNode", "XMLXIncludeEndNode"))
+function(X, FUN, ..., omitNodeTypes = c("XMLXIncludeStartNode", "XMLXIncludeEndNode"), addFinalizer = NA)
 {
-   kids = xmlChildren(X)
+   kids = xmlChildren(X, addFinalizer = addFinalizer)
    if(length(omitNodeTypes))
      kids = kids[! sapply(kids, function(x) any(inherits(x, omitNodeTypes)) )]   
    lapply(kids, FUN, ...)
 }  
 
 xmlSApply.XMLInternalNode =
-function(X, FUN, ..., omitNodeTypes = c("XMLXIncludeStartNode", "XMLXIncludeEndNode"))
+function(X, FUN, ..., omitNodeTypes = c("XMLXIncludeStartNode", "XMLXIncludeEndNode"), addFinalizer = NA)
 {
-   kids = xmlChildren(X)
+   kids = xmlChildren(X, addFinalizer = addFinalizer)
    if(length(omitNodeTypes))
      kids = kids[! sapply(kids, function(x) any(inherits(x, omitNodeTypes)) )]
    sapply(kids, FUN, ...)
 }  
 
 
-
-xmlParent =
-function(x)
- UseMethod("xmlParent")
-
-
-xmlParent.XMLInternalNode =
-function(x)
+xmlSApply.XMLNodeSet =
+function(X, FUN, ..., omitNodeTypes = c("XMLXIncludeStartNode", "XMLXIncludeEndNode"), addFinalizer = NA)
 {
-  .Call("RS_XML_xmlNodeParent", x, PACKAGE = "XML")
+  sapply(X, FUN, ...)
+}
+
+xmlApply.XMLNodeSet =
+function(X, FUN, ..., omitNodeTypes = c("XMLXIncludeStartNode", "XMLXIncludeEndNode"), addFinalizer = NA)
+{
+  lapply(X, FUN, ...)
+}
+
+getChildrenStrings =
+function(node, encoding = getEncoding(node), asVector = TRUE, len = xmlSize(node),
+          addNames = TRUE)
+{
+   encoding = getEncodingREnum(encoding)
+   .Call("R_childStringValues", node, as.integer(len), as.logical(asVector), as.integer(encoding),
+               as.logical(addNames), PACKAGE = "XML")
 }
 
 
+setMethod("xmlParent", "XMLInternalNode",
+function(x, addFinalizer = NA, ...)
+{
+  .Call("RS_XML_xmlNodeParent", x, addFinalizer, PACKAGE = "XML")
+})
+
+
 newXMLDTDNode <-
-function(nodeName, externalID = character(), systemID = character(), doc = NULL)  
+function(nodeName, externalID = character(), systemID = character(), doc = NULL, addFinalizer = NA)  
 {
   if(length(nodeName) > 1 && missing(externalID))
     externalID = nodeName[2]
   if(length(nodeName) > 2 && missing(systemID))
     systemID = nodeName[3]  
 
-  .Call("R_newXMLDtd", doc, as.character(nodeName), as.character(externalID), as.character(systemID), PACKAGE = "XML")
+  .Call("R_newXMLDtd", doc, as.character(nodeName), as.character(externalID), as.character(systemID),
+          addFinalizer, PACKAGE = "XML")
 }
 
 setInternalNamespace =
@@ -309,7 +368,7 @@ function(doc, finalizer)
 {
   fun = NULL
   if(is.logical(finalizer)) {
-    if(!finalizer)
+    if(is.na(finalizer) || !finalizer)
       return()
     else
       fun = NULL
@@ -325,27 +384,70 @@ function(doc, finalizer)
   .Call("R_addXMLInternalDocument_finalizer", doc, fun, PACKAGE = "XML")
 }
 
+HTML_DTDs =
+  c("http://www.w3.org/TR/html4/frameset.dtd",
+    "http://www.w3.org/TR/html4/loose.dtd",
+    "http://www.w3.org/TR/html4/strict.dtd",
+    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd",
+    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd",
+    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd",
+    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"    
+    )
+
+newHTMLDoc =
+function(dtd = "loose", addFinalizer = TRUE, name = character(),
+          node = newXMLNode("html", newXMLNode("head", addFinalizer = FALSE), newXMLNode("body", addFinalizer = FALSE),
+                             addFinalizer = FALSE))
+{
+  if(is.na(dtd) || dtd == "")
+     dtd = ""
+  else if(tolower(dtd) %in% c("html5", "5"))
+     dtd = "5"
+  else {
+     i = grep(dtd, HTML_DTDs)
+     if(length(i)) {
+        if(length(i) > 1)
+           warning("matched multiple DTDs. Using the first")
+        dtd = HTML_DTDs[i[1]]
+     } else
+        dtd = ""
+   } 
+     
+  
+  doc = newXMLDoc(dtd = dtd, isHTML = TRUE, addFinalizer = addFinalizer, node = node)
+  doc
+}
+
 newXMLDoc <-
 #
 # Creates internal C-level libxml object for representing
 # an XML document/tree of nodes.
 #
-function(dtd = NA, namespaces = NULL, addFinalizer = TRUE, name = character(), node = NULL) 
+function(dtd = "", namespaces = NULL, addFinalizer = TRUE, name = character(), node = NULL,
+          isHTML = FALSE) 
 {
-  ans = .Call("R_newXMLDoc", dtd, namespaces, PACKAGE = "XML")
-
+  if(is(dtd, "XMLInternalNode")) {
+    dtdNode = dtd
+    dtd = character()
+  } else
+    dtdNode = NULL
+  
+  ans = .Call("R_newXMLDoc", dtd, namespaces, as.logical(isHTML), PACKAGE = "XML")
+  class(ans) = oldClass(class(ans))
+  
   addDocFinalizer(ans, addFinalizer)
-
+  
   if(length(name))
-    docName(ans) = as.character(name)
+     docName(ans) = as.character(name)
+
+  if(length(dtdNode))
+     addChildren(ans, dtdNode)
 
   if(length(node)) {
-#   if(missing(shallow)) {
-#     shallow = !is.null(as(node, "XMLInternalDocument"))
-#   }
-#    if(!deep) 
-#      .Call("R_insertXMLNode", ans, node, as.integer(NA), FALSE, PACKAGE = "XML")
-    addChildren(ans, node)
+    if(is.character(node))
+       newXMLTextNode(node, addFinalizer = FALSE, parent = doc)
+    else
+       addChildren(ans, node)
   }
 
   ans
@@ -356,7 +458,7 @@ getOption =
 function(name, default = NULL, converter = NULL)
 {
   if(!exists(name, XMLOptions, inherits = FALSE)) 
-    return(default)
+    return(base::getOption(name, default))
 
    ans = get(name, XMLOptions)
 
@@ -395,11 +497,12 @@ function(name, ..., attrs = NULL,
          at = NA,
          cdata = FALSE,
          suppressNamespaceWarning = getOption('suppressXMLNamespaceWarning', FALSE), #  i.e. warn.
-         sibling = NULL, manageMemory = TRUE
-         )
+         sibling = NULL, addFinalizer = NA,
+         noNamespace = length(namespace) == 0 && !missing(namespace),
+         fixNamespaces = c(dummy = TRUE, default = TRUE)
+        )
 {
-
- noNamespace = length(namespace) == 0 && !missing(namespace)
+   # determine whether we know now that there is definitely no namespace.
  
     # make certain we have a character vector for the attributes.
  if(length(attrs)) {
@@ -418,14 +521,22 @@ function(name, ..., attrs = NULL,
 
      # allow the caller to specify the node name as  ns_prefix:name
      # but we have to create it as name and the set the namespace.
-  ns = nsPrefix = character()
+  ns = character()  # the namespace prefix
   name = strsplit(name, ":")[[1]]
   if(length(name) == 2) {
-     ns = nsPrefix = name[1]
+     ns = name[1]
      name = name[2]
      noNamespace = FALSE
   }
 
+ if(is.list(parent)) {
+    if(length(parent) < 1 ||
+        !(is(parent[[1]], "XMLInternalElementNode") || is(parent[[1]], "XMLInternalDocument")))
+        stop("incorrect value for parent")
+    
+    parent = parent[[1]]
+  }
+
    # if there is no doc, but we have a parent which is an XMLInternalDocument, use that.
  if(missing(doc) && !missing(parent) &&
      inherits(parent, "XMLInternalDocument")) {
@@ -436,8 +547,10 @@ function(name, ..., attrs = NULL,
     # Get the doc from the parent node/document.
  if(is.null(doc) && !is.null(parent))  {
    # doc = as(parent, "XMLInternalDocument")
-   doc = if(inherits(parent, "XMLInternalDocument")) parent
-         else .Call("R_getXMLNodeDocument", parent, PACKAGE = "XML")
+   doc = if(inherits(parent, "XMLInternalDocument"))
+            parent
+         else
+            .Call("R_getXMLNodeDocument", parent, PACKAGE = "XML")
  }
 
 
@@ -445,12 +558,12 @@ function(name, ..., attrs = NULL,
 
      # xmlSetProp() routine in R_newXMLNode() handles namespaces on the attribute names, even checking them.
   node <- .Call("R_newXMLNode", as.character(name), character(), character(), doc, namespaceDefinitions, 
-                   as.integer(manageMemory), PACKAGE = "XML")
+                   addFinalizer, PACKAGE = "XML")
 
- if(!is.null(sibling))
-    addSibling(sibling, node, after = as.logical(at))
- else if(!is.null(parent))
-    addChildren(parent, node, at = at)
+  if(!is.null(sibling))
+     addSibling(sibling, node, after = as.logical(at))
+  else if(!is.null(parent))
+     addChildren(parent, node, at = at)
 
 
  if(TRUE) { # Create the name space definitions here rather than in C code.
@@ -458,101 +571,66 @@ function(name, ..., attrs = NULL,
                    function(i) 
                      newNamespace(node, namespaceDefinitions[[i]], names(namespaceDefinitions)[i], set = FALSE)
                    )
-      names(nsDefs) = names(namespaceDefinitions)
-  }  else
+      if(length(namespaceDefinitions))
+         names(nsDefs) = if(length(names(namespaceDefinitions))) names(namespaceDefinitions) else ""
+  } else
       nsDefs = xmlNamespaceDefinitions(node)
 
        # Now that the namespaces are defined, we can define the attributes which _may_ use them.
   addAttributes(node, .attrs = attrs, suppressNamespaceWarning = suppressNamespaceWarning)
- 
 
- if(length(namespace) > 0) {
-#XXX   cat("<newXMLNode> have namespace", length(namespace) , paste(namespace, collapse = ", "), "\n")
-          # a single element with no name so this is the prefix.
-    if(length(namespace) == 1 && length(names(namespace)) == 0 && !(substring(namespace, 1, 3) %in% c("htt", "ftp"))) {
-      if(namespace != "")
-        ns = nsPrefix = namespace
-    } else {
+ if(is(namespace, "XMLNamespaceRef")) {
+    setInternalNamespace(node, namespace)
+ } else if(is.na(noNamespace) || !noNamespace)  {
+    ns = getNodeNamespace(ns, nsDefs, node, namespace, noNamespace, namespaceDefinitions, parent, suppressNamespaceWarning)
+    if(is.null(ns))
+       !.Call("R_setNamespaceFromAncestors", node, PACKAGE = "XML")
+#          .Call("R_getAncestorDefaultNSDef", node, TRUE, PACKAGE = "XML")
+ }
 
-           # we have names and/or more than one element. So these are namespace definitions
-      if(length(names(namespace)) == 0) 
-        names(namespace) <- rep("", length(namespace))
 
-      if(length(namespace) > 1 && !is.na(match(namespace[1], names(namespace)[-1]))) {
-        if(length(ns)) 
-          warning("ignoring first element of namespace and using prefix from node name, ", ns)
-        else {
-          ns = namespace[1]
-          namespace = namespace[-1]
-        }
-      }
-        
-      
 
-      if(length(namespace) > 1 && sum(names(namespace) == "") > 1)
-        warning("more than one namespace to use as the default")
-
-      nsDefs = lapply(seq(along = namespace),
-                      function(i) {
-                        prefix = names(namespace)[i]
-
-                        newNamespace(node, namespace[[i]], prefix)
-                                # Don't set the namespace. This is just a definition/declaration for
-                                # this node, but not necessarily the namespace to use for this node.
-                                # We set this below
-                      })
-      names(nsDefs) = names(namespace)
-    }
-  }
-
-          # Now handle the prefix for this node.
-    if(length(ns)) {
-      i = match(ns, names(nsDefs))
-      if(is.na(i)) {
-         if(!is.null(parent)) 
-           ns = findNamespaceDefinition(node, ns)
-         else {
-            if(is.character(suppressNamespaceWarning))
-               f = get(suppressNamespaceWarning, mode = "function")
-            else if(is.logical(suppressNamespaceWarning) && !suppressNamespaceWarning) {
-               f = warning 
-            } else
-               f = function(...) {}
-                
-            f("cannot find namespace definition because the node is not in the document and there are no local namespace definitions for this node")
-            attr(node, "xml:namespace") = ns
-            ns = NULL
-          }
-          if(!inherits(ns, "XMLNamespaceRef"))
-             ns <- newNamespace(node, ns, "")
-      } else
-         ns <- nsDefs[[i]]
-    } else  {
-      i = match("", names(nsDefs))
-      ns = if(is.na(i)) NULL else nsDefs[[i]]
-
-      if(!noNamespace && length(ns) == 0 && length(parent) > 0) {
-          ns = xmlNamespaceRef(parent)
-      }
-    }
-
- 
      # Here is where we set the namespace for this node.
- if(length(ns) && (inherits(ns, c("XMLNamespaceRef", "XMLNamespaceDeclaration")) || (is.character(ns) && ns != ""))) {
-#XXX   cat("<newXMLNode> set namespace\n")   
+ if(length(ns) && (inherits(ns, c("XMLNamespaceRef", "XMLNamespaceDeclaration")) || (is.character(ns) && ns != ""))) 
      setXMLNamespace( node, ns) # should this be append = FALSE ?
-   }
 
+
+    # Add any children to this node.
  if(length(.children))  {
    if(!is.list(.children))
       .children = list(.children)
-#XXX   cat("<newXMLNode> add children", length(.children), "\n")
-   addChildren(node, kids = .children, cdata = cdata)
+   addChildren(node, kids = .children, cdata = cdata, addFinalizer = addFinalizer)
+ }
+
+ if(any(fixNamespaces)) { # !is.null(parent)) {
+    xmlFixNamespaces(node, fixNamespaces)
+   # fixDummyNS(node, suppressNamespaceWarning)
  }
 
  node
 }
 
+xmlFixNamespaces =
+function(node, fix)
+{
+
+   if(length(fix) == 1)
+      fix = structure(rep(fix, 2), names = c("dummy", "default"))
+
+   if(length(names(fix)) == 0)
+      names(fix)  = c("dummy", "default")
+
+   
+   if(fix["dummy"])
+      xmlApply(node, function(x) .Call("R_fixDummyNS", x, TRUE, PACKAGE = "XML"))
+   if(fix["default"])
+      .Call("R_getAncestorDefaultNSDef", node, TRUE, PACKAGE = "XML")
+}
+
+
+FixDummyNS = 2L
+FixDefaultNS = 4L
+
 
 xmlNamespaceRef =
 function(node)
@@ -621,8 +699,7 @@ function(node,  namespace, append = FALSE)
   else if(is.character(namespace))
 
      namespace = newNamespace(node, namespace)
-
-  else if(!inherits(namespace, c("XMLNamespaceRef", "XMLNamespaceDeclaration")))
+  else if(!is.null(namespace) && !inherits(namespace, c("XMLNamespaceRef", "XMLNamespaceDeclaration")))
     stop("Must provide a namespace definition, a prefix of existing namespace or a reference to a namespace definition")
 
   .Call("R_xmlSetNs", node, namespace, FALSE, PACKAGE = "XML") 
@@ -638,6 +715,16 @@ setAs("XMLNamespaceDefinition", "character",
          structure(from$uri, names = from$id))
 
 
+setGeneric("xmlNamespace<-",
+            function(x, ..., value)
+              standardGeneric("xmlNamespace<-"))
+
+setMethod("xmlNamespace<-", "XMLInternalNode",
+            function(x, ..., value) {
+               setXMLNamespace(x, value)
+               x
+            })
+
 
 setGeneric("xmlNamespaces<-",
             function(x, append = TRUE, set = FALSE, value)
@@ -768,12 +855,13 @@ newXMLTextNode =
   #
   #  cdata allows the caller to specify that the text be 
   #  wrapped in a newXMLCDataNode
-function(text,  parent = NULL, doc = NULL, cdata = FALSE, escapeEntities = is(text, "AsIs"))
+function(text,  parent = NULL, doc = NULL, cdata = FALSE, escapeEntities = is(text, "AsIs"),
+          addFinalizer = NA)
 {
   if(cdata) 
-    return(newXMLCDataNode(text, parent, doc))
+    return(newXMLCDataNode(text, parent, doc, addFinalizer = addFinalizer))
 
-  a = .Call("R_newXMLTextNode", as.character(text), doc, PACKAGE = "XML")
+  a = .Call("R_newXMLTextNode", as.character(text), doc, addFinalizer, PACKAGE = "XML")
   if(escapeEntities)
     setNoEnc(a)
   
@@ -784,28 +872,28 @@ function(text,  parent = NULL, doc = NULL, cdata = FALSE, escapeEntities = is(te
 }
 
 newXMLPINode <-
-function(name,  text,  parent = NULL, doc = NULL, at = NA)
+function(name,  text,  parent = NULL, doc = NULL, at = NA, addFinalizer = NA)
 {
-  a = .Call("R_newXMLPINode", doc, as.character(name), as.character(text), PACKAGE = "XML")
+  a = .Call("R_newXMLPINode", doc, as.character(name), as.character(text), addFinalizer, PACKAGE = "XML")
   if(!is.null(parent))
     addChildren(parent, a, at = at)
   a  
 }
 
 newXMLCDataNode <-
-function(text, parent = NULL, doc = NULL, at = NA, sep = "\n")
+function(text, parent = NULL, doc = NULL, at = NA, sep = "\n", addFinalizer = NA)
 {
   text = paste(as.character(text), collapse = "\n")
-  a = .Call("R_newXMLCDataNode", doc,  text, PACKAGE = "XML")
+  a = .Call("R_newXMLCDataNode", doc,  text, addFinalizer, PACKAGE = "XML")
   if(!is.null(parent))
     addChildren(parent, a, at = at)
   a  
 }
 
 newXMLCommentNode <-
-function(text, parent = NULL, doc = NULL, at = NA) 
+function(text, parent = NULL, doc = NULL, at = NA, addFinalizer = NA) 
 {
-  a = .Call("R_xmlNewComment",  as.character(text), doc, PACKAGE = "XML")
+  a = .Call("R_xmlNewComment",  as.character(text), doc, addFinalizer, PACKAGE = "XML")
   if(!is.null(parent))
     addChildren(parent, a, at = at)
   a  
@@ -817,28 +905,93 @@ function(oldNode, newNode, ...)
   UseMethod("replaceNodes")
 }
 
+replaceNodes.list =
+function(oldNode, newNode, addFinalizer = NA, ...)
+{
+ mapply(replaceNodes, oldNode, newNode, MoreArgs = list(addFinalizer = addFinalizer, ...))
+}
+
 replaceNodes.XMLInternalNode =
-function(oldNode, newNode, ...)
+function(oldNode, newNode, addFinalizer = NA, ...)
 {
   oldNode = as(oldNode, "XMLInternalNode")
   #XXX deal with a list of nodes.
   newNode = as(newNode, "XMLInternalNode")
   
-  .Call("RS_XML_replaceXMLNode", oldNode, newNode, PACKAGE = "XML")
+  .Call("RS_XML_replaceXMLNode", oldNode, newNode, addFinalizer, PACKAGE = "XML")
 }  
 
+#
+if(FALSE) # This is vectorized for no reason
 "[[<-.XMLInternalNode" =
 function(x, i, j, ..., value)
 {
    if(!is.list(value))
      value = list(value)
 
-   addChildren(x, kids = value, at = i)
+  if(is.character(i)) {
+     if(length(names(x)) == 0)
+         k = rep(NA, length(i))
+     else
+         k = match(i, names(x))
+     
+     if(any(is.na(k))) {
+           # create a node with that name and text 
+         value[is.na(k)] = mapply(function(name, val)
+                                    if(is.character(val))
+                                         newXMLNode(name, val)
+                                    else
+                                         val)
+     }
+     i = k
+   }
+
+   replace =  (i <= xmlSize(x))
+
+   if(any(replace)) {
+     replaceNodes(xmlChildren(x)[i[replace]], value[replace])
+     value = value[!replace]
+     i = i[!replace]
+   }
+
+   if(length(i))
+      addChildren(x, kids = value, at = i)
 
    x
 }  
 
 
+
+
+"[[<-.XMLInternalNode" =
+function(x, i, j, ..., value)
+{
+  if(is.character(i)) {
+     if(length(names(x)) == 0)
+         k = NA
+     else
+         k = match(i, names(x))
+     
+     if(is.na(k) && is.character(value) && !inherits(value, "AsIs")) {
+           # create a node with that name and text
+        value = newXMLNode(i, value)
+     }
+     i = k
+   }
+
+   replace = !is.na(i) & (i <= xmlSize(x))
+
+   if(replace) 
+     replaceNodes(xmlChildren(x)[[i]], value)
+   else
+     addChildren(x, kids = list(value), at = i)
+
+   x
+}  
+
+
+
+
 setNoEnc =
 function(node)
 {
@@ -849,8 +1002,9 @@ function(node)
 }
 
 
+
 addChildren.XMLInternalNode =
-addChildren.XMLInternalDocument =
+addChildren.XMLInternalDocument =  
   #
   # XXX need to expand/recycle the at if it is given as a scalar
   # taking into account if the subsequent elements are lists, etc.
@@ -861,14 +1015,17 @@ addChildren.XMLInternalDocument =
   # in kids.  This is not just the length of kids but
   # the number of nodes since some of the elements might be lists.
   #
-function(node, ..., kids = list(...), at = NA, cdata = FALSE)
+function(node, ..., kids = list(...), at = NA, cdata = FALSE, addFinalizer = NA,
+          fixNamespaces = c(dummy = TRUE, default = TRUE))
 {
   kids = unlist(kids, recursive = FALSE)
 
+  removeNodes(kids)[!sapply(kids, is, "character")]
+
   if(length(kids) == 1 && inherits(kids[[1]], "XMLInternalNode") && is.na(at)) {
      .Call("R_insertXMLNode", kids[[1]], node, -1L, FALSE, PACKAGE = "XML")
-     return(node)
-  }
+#     return(node)
+  } else {
 
 # if(all(is.na(at))) {
 #    kids = lapply(kids, as, function(x) if(is.character(x)) newXMLTextNode(x) else as(x, "XMLInternalNode"))
@@ -894,7 +1051,7 @@ function(node, ..., kids = list(...), at = NA, cdata = FALSE)
                i = kids[[j]]
 
                if(is.character(i)) 
-                 i = newXMLTextNode(i, cdata = cdata)
+                 i = newXMLTextNode(i, cdata = cdata, addFinalizer = addFinalizer)
 
                if(!inherits(i, "XMLInternalNode")) #XX is(i, "XMLInternalNode")
                  i = as(i, "XMLInternalNode")
@@ -912,7 +1069,7 @@ function(node, ..., kids = list(...), at = NA, cdata = FALSE)
                   if(xmlSize(node) < at[j])
                     .Call("R_insertXMLNode", i, node, as.integer(NA), FALSE, PACKAGE = "XML")
                   else
-                    .Call("RS_XML_xmlAddSiblingAt", node[[ at[j] ]], i, after, PACKAGE = "XML") # if at = 0, then shove it in before the sibling.
+                    .Call("RS_XML_xmlAddSiblingAt", node[[ at[j] ]], i, after, addFinalizer, PACKAGE = "XML") # if at = 0, then shove it in before the sibling.
                }
             }))
   }
@@ -922,14 +1079,14 @@ function(node, ..., kids = list(...), at = NA, cdata = FALSE)
       
       if(is.list(i)) {  # can't happen now since we unlist()
          for(k in i)
-            addChildren(node, k)
+            addChildren(node, k, addFinalizer = addFinalizer)
       } else {
 
         if(is.null(i))
            next
      
         if(is.character(i)) 
-           i = newXMLTextNode(i, cdata = cdata)
+           i = newXMLTextNode(i, cdata = cdata, addFinalizer = FALSE)
 
 
         if(!inherits(i, "XMLInternalNode")) {
@@ -948,38 +1105,16 @@ function(node, ..., kids = list(...), at = NA, cdata = FALSE)
         }
      }
     }
+  }
+
+  
+  if(!is(node, "XMLInternalDocument") && any(fixNamespaces))
+    xmlFixNamespaces(node, fixNamespaces)
 
   node
 }
 
 
-setGeneric("findXIncludeStartNodes", 
-function(doc, ...)  
-{
-  standardGeneric("findXIncludeStartNodes")
-})
-
-setMethod("findXIncludeStartNodes", "character", 
-function(doc, ...)  
-{
-  findXIncludeStartNodes(xmlParse(doc), ...)
-})
-
-setMethod("findXIncludeStartNodes", "XMLInternalDocument", 
-function(doc, ...)  
-{
-  findXIncludeStartNodes(xmlRoot(doc), ...)
-})
-
-setMethod("findXIncludeStartNodes", "XMLInternalElementNode", 
-function(doc, ...)  
-{
-  nodes = .Call("R_findXIncludeStartNodes", xmlRoot(doc), PACKAGE = "XML")
-  names(nodes) = sapply(nodes, xmlGetAttr, "href", NA)
-  nodes
-})
-
-
 
 addSibling =
 function(node, ..., kids = list(...), after = NA)
@@ -988,7 +1123,7 @@ function(node, ..., kids = list(...), after = NA)
 }
 
 addSibling.XMLInternalNode =
-function(node, ..., kids = list(...), after = TRUE)  
+function(node, ..., kids = list(...), after = TRUE, addFinalizer = NA)  
 {
    #XXX Why add as children?
    if(FALSE && is.na(after))
@@ -996,7 +1131,7 @@ function(node, ..., kids = list(...), after = TRUE)
    else {
      lapply(kids,
             function(x) {
-              .Call("RS_XML_xmlAddSiblingAt", node, x, as.logical(after), PACKAGE = "XML")
+              .Call("RS_XML_xmlAddSiblingAt", node, x, as.logical(after), addFinalizer, PACKAGE = "XML")
             })
    }
   
@@ -1005,23 +1140,43 @@ function(node, ..., kids = list(...), after = TRUE)
 
 
 removeNodes =
+function(node, free = rep(FALSE, length(node)))
+  UseMethod("removeNodes")
+
+removeNodes.default =
+function(node, free = rep(FALSE, length(node)))
+ NULL
+
+removeNodes.list = removeNodes.XMLNodeList = 
 function(node, free = rep(FALSE, length(node)))
 {
-  if(is.list(node))  {
-    if(!all(sapply(node, inherits, "XMLInternalNode")))
-      stop("removeNode only works on internal nodes at present")
+   if(!all(sapply(node, inherits, "XMLInternalNode"))) {
+      warning("removeNode only works on internal nodes at present")
+      return(NULL)
+   }
 
     free = as.logical(free)
     free = rep(free, length = length(node))
-  } else if(!inherits(node, "XMLInternalNode"))
-     stop("removeNode is intended only for internal nodes")
-  else {            
+    .Call("R_removeInternalNode", node, free, PACKAGE = "XML")
+}
+
+removeNodes.XMLNodeSet =
+function(node, free = rep(FALSE, length(node)))
+{
+   removeNodes.list(node, free)
+}
+    
+
+removeNodes.XMLInternalNode =
+function(node, free = rep(FALSE, length(node)))
+{  
     node = list(node)
     free = as.logical(free)
-  }
-     
-  .Call("R_removeInternalNode", node, free, PACKAGE = "XML")
-}  
+    .Call("R_removeInternalNode", node, free, PACKAGE = "XML")
+}
+
+
+
 
 removeChildren =
 function(node, ..., kids = list(...), free = FALSE)
@@ -1221,7 +1376,6 @@ tmp =
 function(node, append = TRUE, suppressNamespaceWarning = getOption('suppressXMLNamespaceWarning', FALSE), value)
 {
    addAttributes(node, .attrs = value, suppressNamespaceWarning = suppressNamespaceWarning, append = append)
-   node
 }
 
 setMethod("xmlAttrs<-", "XMLInternalElementNode", tmp)
@@ -1371,8 +1525,9 @@ function(node, ..., .attrs = NULL, .namespace = FALSE,
   node
 })
 
+#xmlNamespaceDefinitions =  # ??? added this but overrides other S3 generic.
 namespaceDeclarations =
-function(node, ref = FALSE)
+function(node, ref = FALSE, ...)
 {
   .Call("RS_XML_getNsList", node,  as.logical(ref), PACKAGE = "XML")
 }  
@@ -1450,9 +1605,10 @@ function(node, recursive = TRUE, addFinalizer = FALSE, ...)
            })
 
 setMethod("xmlClone", "XMLInternalDocument",           
-function(node, recursive = TRUE, addFinalizer = TRUE, ...)
+function(node, recursive = TRUE, addFinalizer = NA, ...)
 {
   ans = .Call("RS_XML_clone", node, as.logical(recursive), addFinalizer, PACKAGE = "XML")
+
   addDocFinalizer(ans, addFinalizer)
   ans
 })
@@ -1466,58 +1622,6 @@ function(node, recursive = TRUE, addFinalizer = FALSE, ...)
 
 
 
-findXInclude =
-function(x, asNode = FALSE)
-{
-  while(!is.null(x)) {
-    tmp = getSiblingXIncludeStart(x, asNode)
-    if(!is.null(tmp))
-      return(tmp)
-
-     sib = x
-     if(is(sib, "XMLXIncludeStartNode"))
-        return(if(asNode) sib else xmlAttrs(sib))
-
-     x = xmlParent(x)
-  }
-
-  NULL
-}
-
-getSiblingXIncludeStart =
-function(x, asNode = FALSE)
-{
-     sib = x
-     while(!is.null(sib)) {
-       if(inherits(sib, "XMLXIncludeStartNode"))
-         return(if(asNode) sib else xmlAttrs(sib))         
-       sib <- getSibling(sib, FALSE)
-     }
-
-     NULL
-}
-
-
-getNodeLocation =
-function(node)
-{
-   fil = findXInclude(node)
-   if(is.null(fil))
-     fil = docName(node)
-
-   list(file = fil, line = getLineNumber(node))
-}
-
-
-getLineNumber =
-function(node, ...)
-{
-  if(!is(node, "XMLInternalNode"))
-      stop("This must be an C-level/native/internal XML node, i.e. of class 'XMLInternalNode'. Got ", paste(class(node), collapse = ", "))
-
-  .Call("R_getLineNumber", node, PACKAGE = "XML")
-}
-  
 
 
 
@@ -1549,3 +1653,26 @@ function(doc, what)
   } else
      names(defs)[i]
 }
+
+
+"xmlParent<-" =
+ function(x, ..., value) {
+  addChildren(value, ..., kids = list(x))                
+}
+
+
+setOldClass("XMLNamespaceRef")
+setAs("XMLNamespaceRef", "character",
+       function(from) {
+ 	.Call("R_convertXMLNsRef", from, PACKAGE = "XML")
+       })
+
+
+
+
+xmlSearchNs =
+function(node, ns, asPrefix = TRUE, doc = as(node, "XMLInternalDocument"))
+{
+
+ .Call("R_xmlSearchNs", doc, node, as.character(ns), as.logical(asPrefix), PACKAGE = "XML")
+}
