@@ -36,7 +36,7 @@ void RS_XML(ValidationError)(void *ctx, const char *msg, ...);
 
 
 static USER_OBJECT_ convertNode(USER_OBJECT_ ans, xmlNodePtr node, R_XMLSettings *parserSettings);
-static void NodeTraverse(xmlNodePtr doc, USER_OBJECT_ converterFunctions, R_XMLSettings *parserSettings);
+static void NodeTraverse(xmlNodePtr doc, USER_OBJECT_ converterFunctions, R_XMLSettings *parserSettings, int rootFirst);
 
 
 static USER_OBJECT_ makeSchemaReference(xmlSchemaPtr ref);
@@ -113,7 +113,8 @@ RS_XML(ParseTree)(USER_OBJECT_ fileName, USER_OBJECT_ converterFunctions,
 		        USER_OBJECT_ fullNamespaceInfo, USER_OBJECT_ r_encoding,
 		        USER_OBJECT_ useDotNames,
       		         USER_OBJECT_ xinclude, USER_OBJECT_ errorFun,
-                         USER_OBJECT_ manageMemory)
+  	        	  USER_OBJECT_ manageMemory, USER_OBJECT_ r_parserOptions,
+                          USER_OBJECT_ r_rootFirst)
 {
 
   const char *name;
@@ -129,9 +130,17 @@ RS_XML(ParseTree)(USER_OBJECT_ fileName, USER_OBJECT_ converterFunctions,
 
   const char *encoding = NULL;
   int freeName = 0;
+  int parserOptions = 0;
+  int rootFirst = INTEGER(r_rootFirst)[0];
 
-  if(GET_LENGTH(r_encoding))
+  if(GET_LENGTH(r_encoding)) {
       encoding = CHAR(STRING_ELT(r_encoding, 0));
+      if(!encoding[0])
+	  encoding = NULL;
+  }
+
+  if(Rf_length(r_parserOptions))
+     parserOptions = INTEGER(r_parserOptions)[0];
 
   parserSettings.skipBlankLines = LOGICAL_DATA(skipBlankLines)[0];
   parserSettings.converters = converterFunctions;
@@ -189,13 +198,16 @@ RS_XML(ParseTree)(USER_OBJECT_ fileName, USER_OBJECT_ converterFunctions,
 #endif
 
   if(asTextBuffer) {
-      doc = useHTML ? htmlParseDoc(CHAR_TO_XMLCHAR(name), encoding) : xmlParseMemory(name, strlen(name)); 
+      doc = useHTML ? htmlParseDoc(CHAR_TO_XMLCHAR(name), encoding) : 
+	              xmlReadMemory(name, strlen(name), NULL, encoding, parserOptions) ;
+                	  /* xmlParseMemory(name, strlen(name)) */ 
 
       if(doc != NULL) 
          doc->name = (char *) xmlStrdup(CHAR_TO_XMLCHAR("<buffer>"));
 
   } else {
-      doc = useHTML ? htmlParseFile(XMLCHAR_TO_CHAR(name), encoding) : xmlParseFile(name);
+      doc = useHTML ? htmlParseFile(XMLCHAR_TO_CHAR(name), encoding) : 
+	              xmlReadFile(name, encoding, parserOptions) /* xmlParseFile(name) */ ;
   }
 
 #ifdef RS_XML_SET_STRUCTURED_ERROR 
@@ -253,7 +265,7 @@ RS_XML(ParseTree)(USER_OBJECT_ fileName, USER_OBJECT_ converterFunctions,
 #endif
 #endif   
           PROTECT(rdocObj = R_createXMLDocRef(doc));
-	  NodeTraverse(root, converterFunctions, &parserSettings);
+	  NodeTraverse(root, converterFunctions, &parserSettings, rootFirst);
 	  UNPROTECT(1);
       }
       PROTECT(rdoc = NULL_USER_OBJECT);
@@ -287,10 +299,6 @@ RS_XML(ParseTree)(USER_OBJECT_ fileName, USER_OBJECT_ converterFunctions,
      return(R_createXMLDocRef(doc));
   }
 
-#if 0   /* Now garbage collected by R.*/
-  xmlFreeDoc(doc);
-  R_numXMLDocsFreed++;
-#endif
 
   if(!parserSettings.internalNodeReferences) {
      /* Set the class for the document. */
@@ -311,7 +319,7 @@ enum { FILE_ELEMENT_NAME, VERSION_ELEMENT_NAME, CHILDREN_ELEMENT_NAME, NUM_DOC_E
 
 
 void
-NodeTraverse(xmlNodePtr root, USER_OBJECT_ converterFunctions, R_XMLSettings *parserSettings)
+NodeTraverse(xmlNodePtr root, USER_OBJECT_ converterFunctions, R_XMLSettings *parserSettings, int rootFirst)
 {
   xmlNodePtr c, tmp;
     c = root;
@@ -323,11 +331,17 @@ NodeTraverse(xmlNodePtr root, USER_OBJECT_ converterFunctions, R_XMLSettings *pa
 #else
                c->childs;
 #endif
-        if(tmp)
-	    NodeTraverse(tmp, converterFunctions, parserSettings);
-	PROTECT(ref = R_createXMLNodeRef(c, parserSettings->finalize));
-	convertNode(ref, c, parserSettings);
-	UNPROTECT(1);
+       
+        if(!rootFirst && tmp)
+	    NodeTraverse(tmp, converterFunctions, parserSettings, rootFirst);
+
+        PROTECT(ref = R_createXMLNodeRef(c, parserSettings->finalize));
+        convertNode(ref, c, parserSettings);
+    	UNPROTECT(1);
+
+        if(rootFirst && tmp)
+	    NodeTraverse(tmp, converterFunctions, parserSettings, rootFirst);
+
 	c = c->next;
     }
 }
@@ -1256,11 +1270,13 @@ makeSchemaReference(xmlSchemaPtr schema)
 }
 
 
+#ifndef NO_XML_MEMORY_SHOW_ROUTINE
 void
 RS_XML_MemoryShow()
 {
     xmlMemDisplay(stderr);
 }
+#endif
 
 USER_OBJECT_
 RS_XML_setDocumentName(USER_OBJECT_ sdoc, USER_OBJECT_ sname)
@@ -1493,6 +1509,24 @@ R_getDocEncoding(SEXP r_doc)
 }
 
 
+int 
+getTextElementLineNumber(xmlNodePtr node)
+{
+    int val = -1;
+
+    if(node->parent)
+	val = node->parent->line;
+
+    xmlNodePtr prev = node->prev;
+    while(prev) {
+	if(prev->line > 0) {
+	    val = prev->line;
+	    break;
+	}
+	prev = prev->prev;
+    }
+    return(val);
+}
 
 SEXP
 R_getLineNumber(SEXP r_node)
@@ -1503,7 +1537,9 @@ R_getLineNumber(SEXP r_node)
 	return(NEW_INTEGER(0));
     }
 
-    return(ScalarInteger(node->line));
+//    XML_GET_LINE(node)
+    return(ScalarInteger(node->line == 0 ? 
+			 getTextElementLineNumber(node) : node->line));
 }
 
 
