@@ -1,5 +1,4 @@
 /**
-
  The purpose of this file is to provide the C-level facilities
  to create, modify and manage internal XML DOM nodes at the S
  language level. We want to be able to use the interface defined
@@ -33,12 +32,14 @@
 #else
 #include <libxml/parserInternals.h>
 #include <libxml/xmlmemory.h>
+#include <libxml/HTMLtree.h>
 #endif
 
 
 #define R_USE_XML_ENCODING 1
-#include "Utils.h"  /* R_createXMLNodeRef(), Encoding macros. */
+#include "Utils.h"  /* R_createXMLNodeRef, Encoding macros. */
 
+#include "NodeGC.h"
 
 #ifdef USE_OLD_ROOT_CHILD_NAMES
 # define XML_ROOT(n) (n)->childs
@@ -49,7 +50,9 @@
 void incrementDocRef(xmlDocPtr doc);
 int getNodeCount(xmlNodePtr node);
 void incrementDocRefBy(xmlDocPtr doc, int num);
-SEXP R_createXMLNodeRefDirect(xmlNodePtr node, int addFinalizer);
+
+void RS_XML_recursive_unsetListDoc(xmlNodePtr list);
+
 
 /**
  Create a libxml comment node and return it as an S object
@@ -57,7 +60,7 @@ SEXP R_createXMLNodeRefDirect(xmlNodePtr node, int addFinalizer);
 */
 
 USER_OBJECT_
-R_xmlNewComment(USER_OBJECT_ str, USER_OBJECT_ sdoc)
+R_xmlNewComment(USER_OBJECT_ str, USER_OBJECT_ sdoc, USER_OBJECT_ manageMemory)
 {
     xmlNodePtr node;
     xmlDocPtr doc = NULL;
@@ -69,11 +72,11 @@ R_xmlNewComment(USER_OBJECT_ str, USER_OBJECT_ sdoc)
     txt = CHAR_TO_XMLCHAR(CHAR_DEREF(STRING_ELT(str, 0)));
     node =  doc ? xmlNewDocComment(doc, txt) : xmlNewComment(txt);
 
-    return(R_createXMLNodeRef(node));
+    return(R_createXMLNodeRef(node, manageMemory));  
 }
 
 USER_OBJECT_
-R_newXMLTextNode(USER_OBJECT_ value, USER_OBJECT_ sdoc)
+R_newXMLTextNode(USER_OBJECT_ value, USER_OBJECT_ sdoc, SEXP manageMemory)
 {
    xmlNodePtr node;
     xmlDocPtr doc = NULL;
@@ -88,11 +91,11 @@ R_newXMLTextNode(USER_OBJECT_ value, USER_OBJECT_ sdoc)
     else
 	node = xmlNewText(txt);
 	
-    return(R_createXMLNodeRef(node));
+    return(R_createXMLNodeRef(node, manageMemory));
 }
 
 USER_OBJECT_
-R_newXMLCDataNode(USER_OBJECT_ sdoc, USER_OBJECT_ value)
+R_newXMLCDataNode(USER_OBJECT_ sdoc, USER_OBJECT_ value, USER_OBJECT_ manageMemory)
 {
   xmlDocPtr  doc = NULL;
   xmlNodePtr node;
@@ -105,18 +108,17 @@ R_newXMLCDataNode(USER_OBJECT_ sdoc, USER_OBJECT_ value)
 
   node = xmlNewCDataBlock(doc, CHAR_TO_XMLCHAR(tmp), strlen(tmp));
 
-  return(R_createXMLNodeRef(node));
+  return(R_createXMLNodeRef(node, manageMemory));
 }
 
 
 USER_OBJECT_
-R_newXMLPINode(USER_OBJECT_ sdoc, USER_OBJECT_ name, USER_OBJECT_ content)
+R_newXMLPINode(USER_OBJECT_ sdoc, USER_OBJECT_ name, USER_OBJECT_ content, USER_OBJECT_ manageMemory)
 {
-  xmlDocPtr  doc = NULL;
   xmlNodePtr node;
 
   node = xmlNewPI(CHAR_TO_XMLCHAR(CHAR_DEREF(STRING_ELT(name, 0))), CHAR_TO_XMLCHAR(CHAR_DEREF(STRING_ELT(content, 0))));
-  return(R_createXMLNodeRef(node));
+  return( R_createXMLNodeRef(node, manageMemory) ); 
 }
 
 
@@ -128,62 +130,36 @@ R_newXMLNode(USER_OBJECT_ name, USER_OBJECT_ attrs, USER_OBJECT_ nameSpace, USER
    xmlNsPtr ns = NULL;
    xmlNodePtr node;
 
-   if(GET_LENGTH(sdoc))
+   if(GET_LENGTH(sdoc)) {
        doc = (xmlDocPtr) R_ExternalPtrAddr(sdoc);
 
+       if(doc->type != XML_DOCUMENT_NODE && doc->type != XML_HTML_DOCUMENT_NODE) 
+	   doc = doc->doc;
+   }
+
     
    if(GET_LENGTH(nameSpace) > 0) {
        /* Need the default namespace and then also any other */
       CHAR_DEREF(STRING_ELT(nameSpace, 0));
    }
 
-/*   ns = xmlNewNs(xmlGetRootElement(doc), ); */
    node = xmlNewDocNode(doc, ns, CHAR_TO_XMLCHAR(CHAR_DEREF(STRING_ELT(name, 0))), NULL);
 
-
-#if 0
-/* trying to do this in R now. */
-   if((n = GET_LENGTH(nameSpaceDefinitions)) > 0) {
-       /* Need the default namespace and then also any other */
-     USER_OBJECT_ prefixes = GET_NAMES(nameSpaceDefinitions);
-     for(i = 0; i < n ; i++) {
-         xmlNewNs(node, CHAR_DEREF(STRING_ELT(nameSpaceDefinitions, i)), CHAR_DEREF(STRING_ELT(prefixes, i)));
-     }
-   }
-#endif
-
-#if 0
-/* trying to do this in R now. */
-   n = GET_LENGTH(attrs);
-   if(n > 0) {
-       USER_OBJECT_ attrNames = GET_NAMES(attrs);
-       if(GET_LENGTH(attrNames) != n) {
-	   PROBLEM "names of attributes is not the same length of attributes"
-           ERROR;
-       }
-       for(i = 0; i < n ; i++) {
-            xmlSetProp(node, CHAR_TO_XMLCHAR(CHAR_DEREF(STRING_ELT(attrNames, i))),
-			  CHAR_TO_XMLCHAR(CHAR_DEREF(STRING_ELT(attrs, i))));
-       }
-   }
-#endif
-
-
    if(doc && XML_ROOT(doc) == NULL) {
        XML_ROOT(doc) = node;
    }
 
-   return(INTEGER(manageMemory)[0] > 0 ? R_createXMLNodeRef(node) : R_createXMLNodeRefDirect(node, 0));
+   return( R_createXMLNodeRef(node, manageMemory) );
 }
 
 USER_OBJECT_
-RS_XML_getNextSibling(USER_OBJECT_ s_node, USER_OBJECT_ s_prev)
+RS_XML_getNextSibling(USER_OBJECT_ s_node, USER_OBJECT_ s_prev, USER_OBJECT_ manageMemory)
 {
     xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(s_node), ptr;
 
     ptr = LOGICAL(s_prev)[0] ? node->next : node->prev;
 
-    return(ptr ? R_createXMLNodeRef(ptr) : NULL_USER_OBJECT);
+    return(ptr ? R_createXMLNodeRef(ptr, manageMemory) : NULL_USER_OBJECT);
 }
 
 
@@ -219,6 +195,7 @@ RS_XML_setNodeName(USER_OBJECT_ s_node, USER_OBJECT_ s_name)
     return(NULL_USER_OBJECT);
 }
 
+
 #if 0
 int
 removeNodeNamespace(xmlNodePtr node, xmlNsPtr p)
@@ -266,6 +243,23 @@ removeNodeNamespaceByName(xmlNodePtr node, const char * const id)
     return(0);
 }
 
+SEXP
+R_replaceDummyNS(USER_OBJECT_ s_node, USER_OBJECT_ newNS, USER_OBJECT_ prefix)
+{
+    xmlNodePtr node;
+    if(TYPEOF(s_node) != EXTPTRSXP) {
+	PROBLEM "non external pointer passed to R_replaceDummyNS"
+	    ERROR;
+    }
+	
+    node = (xmlNodePtr) R_ExternalPtrAddr(s_node);
+    removeNodeNamespaceByName(node, CHAR(STRING_ELT(prefix, 0)));
+    
+    return(R_xmlSetNs(s_node, newNS, ScalarLogical(0))); 
+//    return(newNS);
+}
+
+
 SEXP 
 RS_XML_removeAllNodeNamespaces(SEXP s_node)
 {
@@ -297,10 +291,10 @@ RS_XML_removeNodeNamespaces(SEXP s_node, SEXP r_ns)
 {
     int i, n;
     xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(s_node);
-    xmlNsPtr p;
     SEXP el, ans;
     const char *prefix;
-    int t = TYPEOF(r_ns);
+//    xmlNsPtr p;
+//    int t = TYPEOF(r_ns);
     n = Rf_length(r_ns);
     PROTECT(ans = allocVector(LGLSXP, n));
 
@@ -394,7 +388,7 @@ RS_XML_getNsList(USER_OBJECT_ s_node, USER_OBJECT_ asRef)
 	PROTECT(names = NEW_CHARACTER(n));    
 	for(i = 0; i < n ; i++, el = el->next) {
 	    if(el->prefix)
-		SET_STRING_ELT(names, i, COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(el->prefix)));
+		SET_STRING_ELT(names, i, ENC_COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(el->prefix)));
 	    SET_VECTOR_ELT(ans, i, R_createXMLNsRef(el));
 	}
     } else {
@@ -403,9 +397,9 @@ RS_XML_getNsList(USER_OBJECT_ s_node, USER_OBJECT_ asRef)
 	PROTECT(names = NEW_CHARACTER(n));    
 	for(i = 0; i < n ; i++, el = el->next) {
 	    if(el->prefix)
-		SET_STRING_ELT(names, i, COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(el->prefix)));
+		SET_STRING_ELT(names, i, ENC_COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(el->prefix)));
 	    if(el->href)
-   	        SET_STRING_ELT(ans, i, COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(el->href)));
+   	        SET_STRING_ELT(ans, i, ENC_COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(el->href)));
 	}
     }
 
@@ -450,8 +444,9 @@ RS_XML_setRootNode(USER_OBJECT_ r_doc, USER_OBJECT_ r_node)
     doc = (xmlDocPtr) R_ExternalPtrAddr(r_doc);
     node = (xmlNodePtr) R_ExternalPtrAddr(r_node);
 
-    if(!node->doc)
-	node->doc = doc;
+    /*  Set the reference counting information. */
+    //if(!node->doc)
+    //    node->doc = doc;
     xmlDocSetRootElement(doc, node);
 
     return(ScalarLogical(TRUE));
@@ -460,7 +455,7 @@ RS_XML_setRootNode(USER_OBJECT_ r_doc, USER_OBJECT_ r_node)
 SEXP
 R_isNodeChildOfAt(SEXP rkid, SEXP rnode, SEXP rat)
 {
-    int i, at;
+    int i=0, at;
     xmlNodePtr kid, node, ptr;
     node = (xmlNodePtr) R_ExternalPtrAddr(rnode);	
     kid = (xmlNodePtr) R_ExternalPtrAddr(rkid);	
@@ -553,14 +548,13 @@ R_insertXMLNode(USER_OBJECT_ node, USER_OBJECT_ parent, USER_OBJECT_ at, USER_OB
 	} else {
 	    tmp = n;
 
-            if(n->_private)
+            if(n->_private) {
 #ifdef R_XML_DEBUG
-  	       fprintf(stderr, "insertXMLNode: %p incrementing document (%p)  %d\n", n, p->doc, *(int *) n->_private);
-	    if(strcmp(n->name, "table") == 0) {
-		fprintf(stderr, "table\n");
-	    }
+               fprintf(stderr, "insertXMLNode: %p to %p, incrementing document (%p)  %d\n", n, p, p->doc, *(int *) n->_private);
 #endif
-	    incrementDocRefBy(p->doc, getNodeCount(n));
+	       if(p->doc)
+ 	           incrementDocRefBy(p->doc, getNodeCount(n));
+            }
 	}
 	check = xmlAddChild(p, tmp);
 
@@ -571,6 +565,7 @@ R_insertXMLNode(USER_OBJECT_ node, USER_OBJECT_ parent, USER_OBJECT_ at, USER_OB
 #endif
 	break;
     case XML_DOCUMENT_NODE:	
+    case XML_HTML_DOCUMENT_NODE:	
 	check = xmlAddChild(p, n);
 	incrementDocRef((xmlDocPtr) p);
 	break;
@@ -602,7 +597,7 @@ R_insertXMLNode(USER_OBJECT_ node, USER_OBJECT_ parent, USER_OBJECT_ at, USER_OB
 }
 
 USER_OBJECT_
-RS_XML_xmlAddSiblingAt(USER_OBJECT_ r_to, USER_OBJECT_ r_node, USER_OBJECT_ r_after)
+RS_XML_xmlAddSiblingAt(USER_OBJECT_ r_to, USER_OBJECT_ r_node, USER_OBJECT_ r_after, USER_OBJECT_ manageMemory)
 {
     xmlNodePtr p, n, ans;
     
@@ -635,11 +630,11 @@ RS_XML_xmlAddSiblingAt(USER_OBJECT_ r_to, USER_OBJECT_ r_node, USER_OBJECT_ r_af
 	p->doc->children = n;
 
     incrementDocRefBy(p->doc, getNodeCount(n));
-    return(R_createXMLNodeRef(ans));
+    return(R_createXMLNodeRef(ans, manageMemory));
 }
 
 USER_OBJECT_
-RS_XML_replaceXMLNode(USER_OBJECT_ r_old, USER_OBJECT_ r_new)
+RS_XML_replaceXMLNode(USER_OBJECT_ r_old, USER_OBJECT_ r_new, USER_OBJECT_ manageMemory)
 {
     xmlNodePtr Old, New, ans;
 
@@ -656,7 +651,7 @@ RS_XML_replaceXMLNode(USER_OBJECT_ r_old, USER_OBJECT_ r_new)
     }
 
     ans = xmlReplaceNode(Old, New);
-    return(R_createXMLNodeRef(ans));  
+    return(R_createXMLNodeRef(ans, manageMemory));  
 }
 
 
@@ -701,7 +696,7 @@ RS_XML_removeChildren(USER_OBJECT_ s_node, USER_OBJECT_ kids, USER_OBJECT_ freeN
 }
 
 USER_OBJECT_
-R_xmlRootNode(USER_OBJECT_ sdoc, USER_OBJECT_ skipDtd)
+R_xmlRootNode(USER_OBJECT_ sdoc, USER_OBJECT_ skipDtd, USER_OBJECT_ manageMemory)
 {
   xmlDocPtr doc = (xmlDocPtr) R_ExternalPtrAddr(sdoc);
   xmlNodePtr node = NULL;
@@ -725,24 +720,33 @@ R_xmlRootNode(USER_OBJECT_ sdoc, USER_OBJECT_ skipDtd)
       return(NULL_USER_OBJECT);
 
 
-  return(R_createXMLNodeRef(node));  
+  return(R_createXMLNodeRef(node, manageMemory));  
 }
 
 
 /**
  Create an S object representing a newly created internal 
  XML document object.
-
  */
 
 int R_numXMLDocs = 0;
 int R_numXMLDocsFreed = 0;
 
 USER_OBJECT_
-R_newXMLDoc(USER_OBJECT_ dtd, USER_OBJECT_ namespaces)
+R_newXMLDoc(USER_OBJECT_ dtd, USER_OBJECT_ namespaces, USER_OBJECT_ isHTML)
 {
   xmlDocPtr doc;
-  doc = xmlNewDoc(CHAR_TO_XMLCHAR("1.0"));
+  if(LOGICAL(isHTML)[0]) {
+      const char *d = (TYPEOF(dtd) == STRSXP && Rf_length(dtd)) ? 
+                                CHAR_DEREF(STRING_ELT(dtd, 0)) : NULL;
+      if(d[0] == '5')
+	  doc = htmlNewDoc("", NULL);
+      else
+	  doc = htmlNewDocNoDtD(d && d[0] ? CHAR_TO_XMLCHAR(d) : NULL, NULL);
+
+  } else
+      doc = xmlNewDoc(CHAR_TO_XMLCHAR("1.0"));
+
   R_numXMLDocs++;
 
   return(R_createXMLDocRef(doc));
@@ -752,7 +756,7 @@ R_newXMLDoc(USER_OBJECT_ dtd, USER_OBJECT_ namespaces)
 
 
 USER_OBJECT_
-R_newXMLDtd(USER_OBJECT_ sdoc, USER_OBJECT_ sdtdName, USER_OBJECT_ sexternalID, USER_OBJECT_ ssysID)
+R_newXMLDtd(USER_OBJECT_ sdoc, USER_OBJECT_ sdtdName, USER_OBJECT_ sexternalID, USER_OBJECT_ ssysID, USER_OBJECT_ manageMemory)
 {
 
     xmlDocPtr doc = NULL;
@@ -781,12 +785,12 @@ R_newXMLDtd(USER_OBJECT_ sdoc, USER_OBJECT_ sdtdName, USER_OBJECT_ sexternalID,
 /* should we do this???
       xmlAddChild((xmlNodePtr) doc, (xmlNodePtr) DTD); 
 */
-    return(R_createXMLNodeRef((xmlNodePtr) node));
+    return(R_createXMLNodeRef((xmlNodePtr) node, manageMemory));
 }
 
 
 /*
-  The append section here is irrelevant!
+
  */
 USER_OBJECT_
 R_xmlSetNs(USER_OBJECT_ s_node, USER_OBJECT_ s_ns, USER_OBJECT_ append)
@@ -810,12 +814,26 @@ R_xmlSetNs(USER_OBJECT_ s_node, USER_OBJECT_ s_ns, USER_OBJECT_ append)
   return(s_ns);
 }
 
+#if 0
+/* remove if the above is sufficient. */
+SEXP
+RS_XML_setNS(SEXP s_node, SEXP r_ns)
+{
+    xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(s_node);
+    xmlNsPtr ns = (xmlNsPtr) R_ExternalPtrAddr(r_ns);
+    xmlSetNS(node, ns);
+    return(NULL_USER_OBJECT);
+}
+#endif
+
+
+static const char *DummyNamespaceHREF = "<dummy>";
 
 USER_OBJECT_
 R_xmlNewNs(USER_OBJECT_ sdoc, USER_OBJECT_ shref, USER_OBJECT_ sprefix)
 {
   xmlNodePtr doc = (xmlNodePtr) R_ExternalPtrAddr(sdoc);
-  const char *href = CHAR_DEREF(STRING_ELT(shref, 0));
+  const char *href = Rf_length(shref) == 0 ? DummyNamespaceHREF : CHAR_DEREF(STRING_ELT(shref, 0));
   const char *prefix = NULL;
   xmlNsPtr ns;
 
@@ -835,7 +853,7 @@ R_xmlNewNs(USER_OBJECT_ sdoc, USER_OBJECT_ shref, USER_OBJECT_ sprefix)
 
 
 USER_OBJECT_
-RS_XML_clone(USER_OBJECT_ obj, USER_OBJECT_ recursive, USER_OBJECT_ addFinalizer)
+RS_XML_clone(USER_OBJECT_ obj, USER_OBJECT_ recursive, USER_OBJECT_ manageMemory)
 {
     if(TYPEOF(obj) != EXTPTRSXP) {
 	PROBLEM  "clone can only be applied to an internal, C-level libxml2 object"
@@ -851,11 +869,11 @@ RS_XML_clone(USER_OBJECT_ obj, USER_OBJECT_ recursive, USER_OBJECT_ addFinalizer
 	xmlNodePtr node, node_ans;
 	node = (xmlNodePtr) R_ExternalPtrAddr(obj);
 	node_ans = xmlCopyNode(node, INTEGER(recursive)[0]);
-	return(R_createXMLNodeRef(node_ans));
+	return(R_createXMLNodeRef(node_ans, manageMemory));
     } else if(R_isInstanceOf(obj, "XMLInternalDocument") || R_isInstanceOf(obj, "XMLInternalDOM")) {
 	xmlDocPtr doc;
 	doc = (xmlDocPtr) R_ExternalPtrAddr(obj);
-	return(R_createXMLDocRef(xmlCopyDoc(doc, INTEGER(recursive)[0])));
+	return(R_createXMLDocRef(xmlCopyDoc(doc, INTEGER(recursive)[0]))); // , manageMemory));
     }
     
     PROBLEM "clone doesn't (yet) understand this internal data type"
@@ -868,63 +886,6 @@ RS_XML_clone(USER_OBJECT_ obj, USER_OBJECT_ recursive, USER_OBJECT_ addFinalizer
 xmlDocPtr currentDoc;
 #endif
 
-int R_XML_MemoryMgrMarker = 1010101011;
-
-void
-initDocRefCounter(xmlDocPtr doc)
-{
-    int *val;
-    if(doc->_private)
-	return;
-
-    doc->_private = calloc(2, sizeof(int));
-    val = (int *) doc->_private;
-    val[1] = R_MEMORY_MANAGER_MARKER;
-}
-
-void 
-incrementDocRefBy(xmlDocPtr doc, int num)
-{
-    int *val;
-    if(!doc || IS_NOT_OUR_DOC_TO_TOUCH(doc))
-	return;
-
-    if(!doc->_private) {
-	initDocRefCounter(doc);
-    }
-
-    val = (int *) doc->_private;
-
-    (*val) += num;
-}
-
-void
-incrementDocRef(xmlDocPtr doc)
-{ 
-  incrementDocRefBy(doc, 1);
-}
-
-#define GET_NODE_COUNT(n) \
-   n->_private ? *((int*) (n)->_private) : 0
-
-
-
-int getNodeCount(xmlNodePtr node)
-{
-  int val = 0;
-  xmlNodePtr p = node->children;
-
-  if(!node || IS_NOT_OUR_NODE_TO_TOUCH(node))
-    return(0);
-
-  val = GET_NODE_COUNT(node);
-  while(p) {
-    val += getNodeCount(p);
-    p = p->next;
-  }
-  return(val);
-}
-
 
 USER_OBJECT_
 R_createXMLDocRef(xmlDocPtr doc)
@@ -943,13 +904,13 @@ R_createXMLDocRef(xmlDocPtr doc)
 
 #ifdef R_XML_DEBUG
   fprintf(stderr, "creating document reference %s %p, count = %d\n", 
-              doc->URL ? doc->URL : "internally created", doc,
+  	      doc->URL ? doc->URL : "internally created", doc,
               * ((int*) doc->_private));
 #endif
 
   PROTECT(ref = R_MakeExternalPtr(doc, Rf_install("XMLInternalDocument"), R_NilValue));
   PROTECT(tmp = NEW_CHARACTER(1));
-  SET_STRING_ELT(tmp, 0, mkChar("XMLInternalDocument"));
+  SET_STRING_ELT(tmp, 0, mkChar( doc->type == XML_HTML_DOCUMENT_NODE ? "HTMLInternalDocument" : "XMLInternalDocument"));
   SET_CLASS(ref, tmp);
   UNPROTECT(2);
   return(ref);
@@ -1010,7 +971,6 @@ R_getXMLNsRef(USER_OBJECT_ r_node)
 }
 
 
-
 const char * 
 R_getInternalNodeClass(xmlElementType type)
 {
@@ -1019,6 +979,9 @@ R_getInternalNodeClass(xmlElementType type)
         case XML_ELEMENT_NODE:
               p = "XMLInternalElementNode";
               break;
+        case XML_ELEMENT_DECL:
+              p = "XMLInternalElementDeclNode";
+              break;
         case XML_TEXT_NODE:
               p = "XMLInternalTextNode";
               break;
@@ -1061,6 +1024,9 @@ R_getInternalNodeClass(xmlElementType type)
         case XML_DOCUMENT_NODE:
               p = "XMLDocumentNode";
               break;
+        case XML_HTML_DOCUMENT_NODE:
+              p = "XMLHTMLDocumentNode";
+              break;
         case XML_DOCUMENT_TYPE_NODE:
               p = "XMLDocumentTypeNode";
               break;
@@ -1077,136 +1043,6 @@ R_getInternalNodeClass(xmlElementType type)
     return(p);
 }
 
-void
-internal_incrementNodeRefCount(xmlNodePtr node)
-{
-    int *val;
-    if(!node || IS_NOT_OUR_NODE_TO_TOUCH(node) || !node->_private)
-	return;
-    val = (int *) node->_private;
-    (*val)++;
-}
-
-SEXP
-R_getXMLRefCount(SEXP rnode)
-{
-    xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(rnode);
-    if(!node || IS_NOT_OUR_NODE_TO_TOUCH(node) || !node->_private)
-	return(R_NilValue);
-    return(ScalarInteger(*((int *) node->_private)));
-}
-
-int
-checkDescendantsInR(xmlNodePtr node, int process)
-{
-    xmlNodePtr p;
-    if(!node && (process || IS_NOT_OUR_NODE_TO_TOUCH(node)))
-	return(0);
-
-    if(node->_private)
-	return(1);
-
-    p = node->children;
-    while(p) {
-	if(checkDescendantsInR(p, 0))
-	    return(1);
-	p = p->next;
-    }
-    return(0);
-}
-
-void
-internal_decrementNodeRefCount(xmlNodePtr node)
-{
-    int *val;
-       /* */    
-    if(!node || IS_NOT_OUR_NODE_TO_TOUCH(node))
-                                 /* if node->_private == NULL, should
-				  * we free this node?, i.e. if it is
-				  * not in a parent or a document.
-                                  No! Basically we shouldn't get here
-                                  if we have not set the _private. We
-                                  set the finalizer having set the _private */
-	return;
-
-    if(!node->_private)
-	return;
-
-
-    /*  Get the value of the reference count and decrement it by 1.
-        If we are now at 0, then we can potentially free this node.
-        Certainly, if we are at 0, we should remove the reference
-        count memory altogether.
-        Now that _we_ no longer need the node, perhaps we can free it.
-        But we have to make certain that we don't free it if
-         a) it is a child of another node or 
-         b) if it is within a document and that document is still "in  play".
-              To determine if the document is "in play" we look at it's
-              reference count. 
-              We decrement it by one since we added one to it for this
-              node.
-              If that makes the document's reference count 0, then we
-              free it.
-      
-     */
-    val = (int *) node->_private;
-    (*val)--;
-#ifdef R_XML_DEBUG
-    fprintf(stderr, "decremented node (%s, %d) to %d  (%p)\n", node->name, node->type, *val, node);fflush(stderr);
-#endif
-    if(*val == 0) {
-
-	free(node->_private);
-        node->_private = NULL;
-	if(node->doc && !IS_NOT_OUR_DOC_TO_TOUCH(node->doc)) {
-	    val = (int *) node->doc->_private;
-	    if(val) (*val)--;
-	    if(!val || *val == 0) {
-		/* Consolidate with R_xmlFreeDoc */
-#ifdef R_XML_DEBUG
-		fprintf(stderr, "releasing document (for node) %p %s (%s)\n", node->doc, node->doc->URL ? node->doc->URL : "?", val ? "has zero count" : "no count");fflush(stderr);
-#endif
-		if(val) free(node->doc->_private);
-		node->doc->_private = NULL;
-/*		fprintf(stderr, "<r:freeDoc addr='%p'/>\n", node->doc); */
-		xmlFreeDoc(node->doc);
-		R_numXMLDocsFreed++;
-	    }
-	} else if(!node->parent) {
-            int hold;
-	    hold = checkDescendantsInR(node, 1);
-            if(!hold)
-   	       xmlFreeNode(node);
-	} else {
-            /* So we have a parent.  But what if that parent is not
-               being held as an R variable. We need to free the node.
-               We need to make this smarter to see what parts of the 
-               tree we can remove.  For instance, we might be holding
-               onto this one, but not the parent, but that parent has
-               a second child which is being held onto.
-
-               So we go to the top of the node tree and check for its descendants
-            */
-	    int hold;
-            xmlNodePtr p = node->parent;
-	    while(p->parent)  p = p->parent;
-
-	    hold = checkDescendantsInR(p, 0);
-	    if(!hold)
-		xmlFree(p);
-        
-	}
-    }
-}
-
-
-
-void
-decrementNodeRefCount(SEXP rnode)
-{
-    xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(rnode);
-    internal_decrementNodeRefCount(node);
-}
 
 
 SEXP
@@ -1215,13 +1051,20 @@ R_createXMLNodeRefDirect(xmlNodePtr node, int addFinalizer)
   SEXP ref, tmp;
 
   PROTECT(ref = R_MakeExternalPtr(node, Rf_install("XMLInternalNode"), R_NilValue));
+
 #ifdef XML_REF_COUNT_NODES
-#ifdef XML_WITH_DEBUG
+
+  if(addFinalizer > 0 || (addFinalizer < 0 && !IS_NOT_OUR_NODE_TO_TOUCH(node))) {
+#ifdef R_XML_DEBUG
 fprintf(stderr, "Creating reference with finalizer for %s (%p) '%s'\n", 
             node->name, node, node->type == XML_TEXT_NODE ? node->content : "");fflush(stderr);
 #endif
-  if(addFinalizer > 0 || (addFinalizer < 0 && !IS_NOT_OUR_NODE_TO_TOUCH(node)))
      R_RegisterCFinalizer(ref, decrementNodeRefCount);
+  }
+/*
+#else
+#warning "no ref counting enabled"
+*/
 #endif
   PROTECT(tmp = NEW_CHARACTER(3));
   SET_STRING_ELT(tmp, 0, mkChar(R_getInternalNodeClass(node->type)));
@@ -1232,56 +1075,28 @@ fprintf(stderr, "Creating reference with finalizer for %s (%p) '%s'\n",
   return(ref);
 }
 
-int
-clearNodeMemoryManagement(xmlNodePtr node)
-{
-    xmlNodePtr tmp;
-    int ctr = 0;
-    if(node->_private) {
-	free(node->_private);
-	node->_private = NULL;
-	ctr++;
-    }
-
-    tmp = node->children;
-    while(tmp) {
-	tmp = tmp->next;
-        if(tmp)
-   	   ctr += clearNodeMemoryManagement(tmp);
-    }
-    return(ctr);
-}
-
-SEXP
-R_clearNodeMemoryManagement(SEXP r_node)
-{
-   xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(r_node);
-   int val;
-
-   if(!node) 
-       return(ScalarInteger(-1));
-
-   val = clearNodeMemoryManagement(node);
-
-   return(ScalarInteger(val));
-}
 
 
 
 /**
-
+Used to be used as 
+R_XML_getManageMemory(manageMemory, node->doc, node) > 0 ? R_createXMLNodeRef() : R_createXMLNodeRefDirect(node, 0)); 
  */
 USER_OBJECT_
-R_createXMLNodeRef(xmlNodePtr node)
+R_createXMLNodeRef(xmlNodePtr node, USER_OBJECT_ finalize)
 {
   int *val;
+  int addFinalizer = 0;
 
   if(!node)
       return(NULL_USER_OBJECT);
 
+
+  addFinalizer = R_XML_getManageMemory(finalize, node->doc, node);
+
 /*  !IS_NOT_OUR_NODE_TO_TOUCH(node) */
-  if((node->_private && ((int*)node->_private)[1] == (int) R_MEMORY_MANAGER_MARKER)
-     || !node->doc || (!(IS_NOT_OUR_DOC_TO_TOUCH(node->doc)))) {
+  if(addFinalizer && ((node->_private && ((int*)node->_private)[1] == (int) R_MEMORY_MANAGER_MARKER)
+		      || !node->doc || (!(IS_NOT_OUR_DOC_TO_TOUCH(node->doc))))) {
       if(node->_private == NULL) {
         node->_private = calloc(2, sizeof(int));
 	val = (int *) node->_private;
@@ -1297,7 +1112,7 @@ R_createXMLNodeRef(xmlNodePtr node)
 #endif
   }
 
-  return(R_createXMLNodeRefDirect(node, !IS_NOT_OUR_NODE_TO_TOUCH(node)));
+  return(R_createXMLNodeRefDirect(node, addFinalizer /* !IS_NOT_OUR_NODE_TO_TOUCH(node) */ ));
 }
 
 
@@ -1309,8 +1124,8 @@ R_createXMLNodeRef(xmlNodePtr node)
 SEXP
 R_addXMLNodeFinalizer(SEXP r_node)
 {
-   xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(r_node);
 #ifdef XML_REF_COUNT_NODES /* ??? should this be ifndef or ifdef.??  */
+   xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(r_node);
   R_RegisterCFinalizer(r_node, decrementNodeRefCount);
 #endif
   return(r_node);
@@ -1430,7 +1245,7 @@ R_saveXMLDOM(USER_OBJECT_ sdoc, USER_OBJECT_ sfileName, USER_OBJECT_ compression
 
 	if(mem) {
 	    DECL_ENCODING_FROM_DOC(doc)
-	    SET_STRING_ELT(ans, 0, COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(mem))); 
+	    SET_STRING_ELT(ans, 0, ENC_COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(mem))); 
 	    xmlFree(mem);
 	} else { 
                /*XXX get the error message from libxml2 */
@@ -1465,8 +1280,59 @@ RS_XML_setDoc(USER_OBJECT_ snode, USER_OBJECT_ sdoc)
     return(R_createXMLDocRef(doc));
 }
 
+#if 0
+void
+RS_XML_recursive_unsetDoc(xmlNodePtr node)
+{
+    xmlNodePtr tmp;
+    node->doc = NULL;
+    tmp = node->children;
+    while(tmp) {
+	RS_XML_recursive_unsetDoc(tmp);
+	tmp = tmp->next;
+    }
+}
+#endif
+
+/* The following two routines are from Paul Murrell.
+   They fix a problem with xpathApply() changing the document
+   presumably when doing an XPath query on a node within a document.
+   The old version didn't deal with the properties on the node.
+ */
+void
+RS_XML_recursive_unsetTreeDoc(xmlNodePtr node) {
+    xmlAttrPtr prop;
+
+    if (node == NULL)
+	return;
+    if(node->type == XML_ELEMENT_NODE) {
+        prop = node->properties;
+        while (prop != NULL) {
+            prop->doc = NULL;
+            RS_XML_recursive_unsetListDoc(prop->children);
+            prop = prop->next;
+        }
+    }
+    if (node->children != NULL)
+        RS_XML_recursive_unsetListDoc(node->children);
+    node->doc = NULL;
+}
+
+void
+RS_XML_recursive_unsetListDoc(xmlNodePtr list) {
+    xmlNodePtr cur;
+
+    if (list == NULL)
+	return;
+    cur = list;
+    while (cur != NULL) {
+        RS_XML_recursive_unsetTreeDoc(cur);
+	cur = cur->next;
+    }
+}
+
 USER_OBJECT_
-RS_XML_unsetDoc(USER_OBJECT_ snode, USER_OBJECT_ unlink, USER_OBJECT_ r_parent)
+RS_XML_unsetDoc(USER_OBJECT_ snode, USER_OBJECT_ unlink, USER_OBJECT_ r_parent, USER_OBJECT_ recursive)
 {
     xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(snode);
     if(!node) {
@@ -1488,6 +1354,10 @@ RS_XML_unsetDoc(USER_OBJECT_ snode, USER_OBJECT_ unlink, USER_OBJECT_ r_parent)
         node->parent = (xmlNodePtr) R_ExternalPtrAddr(snode);
     }
 
+    if(LOGICAL(recursive)[0]) {
+	RS_XML_recursive_unsetTreeDoc(node);
+    }
+
     return(ScalarLogical(TRUE));
 }
 
@@ -1551,7 +1421,8 @@ Test:
   .Call("RS_XML_printXMLNode", a, as.integer(1), as.integer(1), character())
 */
 USER_OBJECT_
-RS_XML_printXMLNode(USER_OBJECT_ r_node, USER_OBJECT_ level, USER_OBJECT_ format, USER_OBJECT_ indent, USER_OBJECT_ r_encoding)
+RS_XML_printXMLNode(USER_OBJECT_ r_node, USER_OBJECT_ level, USER_OBJECT_ format, 
+		    USER_OBJECT_ indent, USER_OBJECT_ r_encoding, USER_OBJECT_ r_encoding_int)
 {
     USER_OBJECT_ ans;
     xmlNodePtr node;
@@ -1581,12 +1452,10 @@ RS_XML_printXMLNode(USER_OBJECT_ r_node, USER_OBJECT_ level, USER_OBJECT_ format
 
     if(xbuf->use > 0) {
         /*XXX this const char * in CHARSXP means we have to make multiple copies. */
-#if 0
-	char *rbuf = R_alloc(sizeof(char) * (xbuf->use + 1));
-	memcpy(rbuf, xbuf->content, xbuf->use + 1);
-	PROTECT(tmp = mkChar(rbuf));
-#endif
-	ans = ScalarString(mkChar(xbuf->content));
+     if(INTEGER(r_encoding_int)[0] == CE_NATIVE)
+        ans = ScalarString(CreateCharSexpWithEncoding(encoding, xbuf->content));
+     else
+        ans = ScalarString(mkCharCE(xbuf->content, INTEGER(r_encoding_int)[0]));
     } else
       ans = NEW_CHARACTER(1);
 
@@ -1612,7 +1481,7 @@ SEXP
 R_setXMLInternalTextNode_value(SEXP node, SEXP value)
 {
    xmlNodePtr n = (xmlNodePtr) R_ExternalPtrAddr(node);
-   xmlChar *tmp;
+//   xmlChar *tmp;
    const char *str;
 
    DECL_ENCODING_FROM_NODE(n)
@@ -1621,17 +1490,21 @@ R_setXMLInternalTextNode_value(SEXP node, SEXP value)
        PROBLEM  "Can only set value on an text node"
 	   ERROR;
    }
-    
-   if(n->content)
-       xmlFree(n->content);
 
    str = CHAR(STRING_ELT(value, 0));
-   n->content = xmlCharStrndup(str, strlen(str));
-/*xmlStrdup(CHAR_TO_XMLCHAR(CHAR(STRING_ELT(value, 0))));*/
+   xmlNodeSetContent(n, str);
     
    return(node);
 }
 
+SEXP 
+R_xmlSetContent(SEXP node, SEXP content)
+{
+    xmlNodePtr n = (xmlNodePtr) R_ExternalPtrAddr(node);
+    xmlNodeSetContent(n, CHAR_TO_XMLCHAR(CHAR_DEREF(STRING_ELT(content, 0))));
+    return(R_NilValue);
+}
+
 SEXP
 R_xmlNodeValue(SEXP node, SEXP raw, SEXP r_encoding)
 {
@@ -1657,11 +1530,11 @@ R_xmlNodeValue(SEXP node, SEXP raw, SEXP r_encoding)
    }
 */
    if(tmp) {
-#if 0
-     ans = ScalarString(CreateCharSexpWithEncoding(encoding, tmp));
-#else
-     ans = ScalarString(mkCharCE(tmp, INTEGER(r_encoding)[0]));
-#endif
+     if(INTEGER(r_encoding)[0] == CE_NATIVE)
+        ans = ScalarString(CreateCharSexpWithEncoding(encoding, tmp));
+     else
+        ans = ScalarString(mkCharCE(tmp, INTEGER(r_encoding)[0]));
+
 
      free(tmp);
 //     ans = mkString(XMLCHAR_TO_CHAR(tmp));
@@ -1690,9 +1563,9 @@ R_xmlNsAsCharacter(USER_OBJECT_ s_ns)
   SET_STRING_ELT(names, 1, mkChar("href"));
 
   if(ns->prefix)
-      SET_STRING_ELT(ans, 0, COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(ns->prefix)));
+      SET_STRING_ELT(ans, 0, ENC_COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(ns->prefix)));
   if(ns->href)
-      SET_STRING_ELT(ans, 1, COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(ns->href)));
+      SET_STRING_ELT(ans, 1, ENC_COPY_TO_USER_STRING(XMLCHAR_TO_CHAR(ns->href)));
 
   SET_NAMES(ans, names);
   UNPROTECT(2);
@@ -1723,9 +1596,13 @@ RS_XML_isDescendantOf(USER_OBJECT_ r_node, USER_OBJECT_ r_top, USER_OBJECT_ stri
 	    ERROR;
     }
 
+/*XXX */
+    if(node->type == XML_NAMESPACE_DECL)
+	return(ScalarLogical(TRUE));
+
     ptr = node;
 
-    while(ptr && ptr->type != XML_DOCUMENT_NODE) {
+    while(ptr && ptr->type != XML_DOCUMENT_NODE && ptr->type != XML_HTML_DOCUMENT_NODE) {
 	if(ptr == top)
 	    return(ScalarLogical(ptr == node && LOGICAL(strict)[0] ? FALSE : TRUE));
 	ptr = ptr->parent;
@@ -1738,7 +1615,7 @@ RS_XML_isDescendantOf(USER_OBJECT_ r_node, USER_OBJECT_ r_top, USER_OBJECT_ stri
 SEXP
 R_XML_indexOfChild(SEXP r_node)
 {
-    xmlNodePtr node, parent, ptr;
+    xmlNodePtr node, ptr; // parent
     int i = 0;
     node = (xmlNodePtr) R_ExternalPtrAddr(r_node);    
     ptr = node->parent->children;
@@ -1753,3 +1630,178 @@ R_XML_indexOfChild(SEXP r_node)
 
     return(R_NilValue);
 }
+
+
+
+SEXP
+R_setNamespaceFromAncestors(SEXP r_node)
+{
+    xmlNodePtr node, ptr;
+    node = (xmlNodePtr) R_ExternalPtrAddr(r_node);    
+    ptr = node->parent;
+    while(ptr) {
+	if((ptr->type != XML_HTML_DOCUMENT_NODE && ptr->type != XML_DOCUMENT_NODE) && 
+              ptr->ns && ptr->ns->href && (!ptr->ns->prefix || !ptr->ns->prefix[0])) {
+	    xmlSetNs(node, ptr->ns);
+	    return(ScalarLogical(TRUE));
+	}
+	ptr = ptr->parent;
+    }
+    return(ScalarLogical(FALSE));
+}
+
+
+#ifdef R_HAS_REMOVE_FINALIZERS
+int
+xmlNode_removeFinalizers(xmlNodePtr node)
+{
+  xmlNodePtr tmp;
+  int count = 0;
+
+#if R_XML_DEBUG
+fprintf(stderr, "xml removeFinalizers  %p %s\n", node, node->name);
+#endif
+  count = R_RemoveExtPtrWeakRef_direct(node);
+
+  tmp = node->children;
+  while(tmp) {
+      count += xmlNode_removeFinalizers(tmp);
+      tmp = tmp->next;
+  }
+  return(count);
+}
+
+SEXP
+R_xmlNode_removeFinalizers(SEXP r_node)
+{
+    int num;
+    xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(r_node);    
+    num = xmlNode_removeFinalizers(node);
+    return(ScalarInteger(num));
+}
+#endif
+
+
+
+
+
+
+SEXP
+R_xmlSearchNs(SEXP r_doc, SEXP r_node, SEXP r_ns, SEXP r_asPrefix)
+{
+    const char * val;
+    xmlNsPtr ns;
+
+    xmlDocPtr doc = (xmlDocPtr) r_doc == NULL_USER_OBJECT ? NULL : R_ExternalPtrAddr(r_doc);    
+    xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(r_node);    
+
+    if(Rf_length(r_ns) == 0)
+	return(NEW_CHARACTER(0));
+
+    val = CHAR_DEREF(STRING_ELT(r_ns, 0));
+    
+    ns = LOGICAL(r_asPrefix)[0] ? xmlSearchNs(doc, node, val) : xmlSearchNsByHref(doc, node, val);
+
+    if(!ns)
+	return(NEW_CHARACTER(0));
+    else {
+	SEXP r_ans;
+	PROTECT(r_ans =  mkString(ns->href));
+	SET_NAMES(r_ans, mkString(ns->prefix ? XMLCHAR_TO_CHAR(ns->prefix) : ""));
+	UNPROTECT(1);
+	return(r_ans);
+    }
+}
+
+
+USER_OBJECT_
+R_getChildByIndex(USER_OBJECT_ r_node, USER_OBJECT_ r_index, USER_OBJECT_ r_addFinalizer)
+{
+    xmlNodePtr node, parent, ptr;
+    int i = 0, idx;
+    node = (xmlNodePtr) R_ExternalPtrAddr(r_node);    
+    ptr = node->children;
+    idx = INTEGER(r_index)[0];
+
+    while(ptr && i < idx) {
+	ptr = ptr->next;
+	i++;
+    }
+
+    return(R_createXMLNodeRef(ptr, r_addFinalizer));
+}
+
+
+USER_OBJECT_
+R_getChildByName(USER_OBJECT_ r_node, USER_OBJECT_ r_index, USER_OBJECT_ r_addFinalizer)
+{
+    xmlNodePtr node, parent, ptr;
+    int i = 0, idx;
+    node = (xmlNodePtr) R_ExternalPtrAddr(r_node);    
+    ptr = node->children;
+    const char *name = CHAR_DEREF(STRING_ELT(r_index, 0));
+
+    while(ptr) {
+	if(ptr->name && strcmp(name, ptr->name) == 0)
+	    break;
+	ptr = ptr->next;
+    }
+
+    return(R_createXMLNodeRef(ptr, r_addFinalizer));
+}
+
+
+/*
+ This is a C-level version equivalent to
+     xmlApply(node, xmlValue)
+
+ 
+*/
+
+USER_OBJECT_
+R_childStringValues(SEXP r_node, SEXP r_len, SEXP r_asVector, SEXP r_encoding, SEXP r_addNames)
+{
+    xmlNodePtr node, kid;
+    int len, i;
+    SEXP ans, names = NULL;
+    int asVector = LOGICAL(r_asVector)[0];
+    int encoding = INTEGER(r_encoding)[0];
+    xmlChar *tmp;
+    int nprotect = 0;
+
+    node = (xmlNodePtr) R_ExternalPtrAddr(r_node);    
+    len = INTEGER(r_len)[0];
+
+    if(asVector)
+	ans = NEW_CHARACTER(len);
+    else
+	ans = NEW_LIST(len);
+
+    PROTECT(ans); nprotect++;
+
+    if(LOGICAL(r_addNames)[0]) {
+	PROTECT(names = NEW_CHARACTER(len));
+	nprotect++;
+    }
+
+
+    for(i = 0, kid = node->children; kid && i < len; i++, kid = kid->next) {
+	tmp  = xmlNodeGetContent(kid);	
+	SEXP val = mkCharCE(tmp, encoding);
+	PROTECT(val);
+	if(asVector)
+	    SET_STRING_ELT(ans, i, val);
+	else
+	    SET_VECTOR_ELT(ans, i, ScalarString(val));
+	if(names && kid->name) {
+	    SET_STRING_ELT(names, i, mkCharCE(kid->name, encoding));
+	}
+	UNPROTECT(1);
+    }
+
+    if(names)
+	SET_NAMES(ans, names);
+
+    UNPROTECT(nprotect);
+    return(ans);
+}
